#define MICHLIB_NOSOURCE
#include "layereddataz.h"

MString LayeredDataZ::Info() const
{
 if(!isOk()) return "";
 MString d;
 for(size_t i = 0; i < NDepths(); i++) d += MString(" ") + "(" + i + " " + Depth(i) + ")";

 std::set<MString> vars;
 for(const auto& f: nc) GetVars(f, vars);

 MString svars;
 {
  bool first = true;
  for(const auto& v: vars)
  {
   svars += (first ? "" : ", ") + v;
   first = false;
  }
 }

 // clang-format off
 return
  "Dataset: " + Title() + "\n" +
  " Begin date: " + Time(0).ToString() + "\n" +
  " End date: " + Time(NTimes()-1).ToString() + "\n" +
  " Time step: " + Timestep() + " seconds\n" +
  " Time moments: " + NTimes() + "\n" +
  " Region: (" + lonb + " : " + lone + ") x (" + latb + " : " + late + ")\n" +
  " Grid: " + dname.nx + "x" + dname.ny + " (" + lonstep + " x " + latstep + ")\n" +
  " Depths:" + d + "\n" +
  " Supported variables: " + svars;
 // clang-format on
}

MString LayeredDataZ::Open(const MString& dataset)
{
 nc.clear();
 MString proxyurl = GPL.ParameterSValue("USEPROXY", "");
 if(proxyurl.Exist()) proxy.Activate("all_proxy", proxyurl);

 nc.clear();
 size_t i = 1;
 while(true)
 {
  MString url = GPL.ParameterSValue(dataset + "_URL" + i, "");
  if(url.Exist())
  {
   // Split url on product and dataset
   auto words = url.Split(":");
   if(words.size() == 0 || words.size() > 2)
   {
    nc.clear();
    return "Invalid url " + url;
   }
   MString product = words[0];
   MString dataset = words.size() == 2 ? words[1] : "";

   nc.emplace_back();
   {
    auto ret = nc.back().Open(product, dataset);
    if(!ret)
    {
     nc.clear();
     return "Can't open " + dataset + " of " + product;
    }
   }
  }
  else
   break;
  i++;
 }
 if(nc.size() == 0) return "No urls for dataset " + dataset + " specified in config";

 dname = GetDNames(nc[0]);
 if(!(dname.lonname.Exist() && dname.latname.Exist()))
 {
  nc.clear();
  return "Can't find longitude/latitude";
 }
 if(!dname.timename.Exist())
 {
  nc.clear();
  return "Can't find time";
 }

 auto cn = GetCNames(nc[0]);

 // Read times
 for(auto& f: nc)
 {
  auto ret = f.ReadTimes(cn.timename);
  if(!ret)
  {
   nc.clear();
   return "Can't read times";
  }
  times.insert(times.end(), f.Times().begin(), f.Times().end());
 }
 std::sort(times.begin(), times.end());
 auto last = std::unique(times.begin(), times.end());
 times.erase(last, times.end());

 depthinv = false;
 if(cn.depthname.Exist())
 {
  auto ret = nc[0].Read(cn.depthname, depths);
  if(!ret)
  {
   nc.clear();
   return "Can't read depths";
  }
  if(depths.back() <= 0 && depths.front() <= 0) std::ranges::transform(depths, depths.begin(), std::negate{});
  if(depths.back() < depths.front() && depths.size() > 1)
  {
   depthinv = true;
   for(size_t i = 0; i < depths.size() - i - 1; i++) std::swap(depths[i], depths[depths.size() - i - 1]);
  }
 }
 else // Surface only data
 {
  depths.resize(1);
  depths[0] = 0;
 }

 std::vector<double> lons, lats;
 {
  auto ret = nc[0].Read(cn.lonname, lons);
  if(!ret)
  {
   nc.clear();
   return "Can't get longitudes";
  }
 }
 {
  auto ret = nc[0].Read(cn.latname, lats);
  if(!ret)
  {
   nc.clear();
   return "Can't get latitudes";
  }
 }

 lonb    = lons[0];
 latb    = lats[0];
 lone    = lons.back();
 late    = lats.back();
 lonstep = (lone - lonb) / (dname.nx - 1);
 latstep = (late - latb) / (dname.ny - 1);

 return "";
}

std::pair<const BaseParameters*, MString> LayeredDataZ::Parameters(michlib_internal::ParameterListEx& pars, const CLArgs& args, const struct Region& reg) const
{
 std::unique_ptr<struct Parameters> ppar{new struct Parameters};

 ppar->layer = args.contains("layer") ? args.at("layer").ToInteger<size_t>() : 0;
 if(!args.contains("depth") && ppar->layer >= NDepths()) return {nullptr, MString("Layer ") + ppar->layer + " is too deep!"};
 real depth = args.contains("depth") ? args.at("depth").ToReal() : Depth(ppar->layer);

 {
  auto dom    = DetGeoDomain(lonb, lone);
  real lon1   = ToGeoDomain(reg.lonb, dom);
  real lon2   = ToGeoDomain(reg.lone, dom);
  real lat1   = reg.latb;
  real lat2   = reg.late;
  bool global = lone - lonb + 1.5 * lonstep > 360.0;

  // Special case when the longitude lies in a small sector between the end and the start
  if(global)
  {
   if(lon1 < lonb) lon1 = lone;
   if(lon2 > lone) lon2 = lonb;
  }
  else
  {
   if(lon1 < lonb) lon1 = lonb;
   if(lon2 > lone) lon2 = lone;
  }

  ppar->xb = static_cast<size_t>(Floor((lon1 - lonb) / lonstep));
  ppar->xe = static_cast<size_t>(Ceil((lon2 - lonb) / lonstep));

  if(ppar->xb == ppar->xe) return {nullptr, "Lonb must be not equal late"};
  if(!global && ppar->xb > ppar->xe) return {nullptr, "Lonb must be lesser then lone"};

  ppar->yb = static_cast<size_t>(Floor((lat1 - latb) / latstep));
  ppar->ye = static_cast<size_t>(Ceil((lat2 - latb) / latstep));
  if(ppar->ye > dname.ny - 1) ppar->ye = dname.ny - 1;
  if(ppar->yb >= ppar->ye) return {nullptr, "Latb must be lesser then late"};

  if(depth < 0.0 || depth > depths.back())
   ppar->layer = (depth < 0.0) ? 0 : (depths.size() - 1);
  else
   for(size_t i = 0; i < depths.size() - 1; i++)
   {
    if(depth >= depths[i] && depth <= depths[i + 1])
    {
     ppar->layer = (depth - depths[i] <= depths[i + 1] - depth) ? i : (i + 1);
     break;
    }
   }
 }

 pars.SetParameter("depth", Depth(ppar->layer));
 pars.SetParameter("layer", ppar->layer);
 pars.SetParameter("dataset", Title());
 pars.SetParameter("lonb", Lon(ppar->xb));
 pars.SetParameter("latb", Lat(ppar->yb));
 pars.SetParameter("lone", Lon(ppar->xe));
 pars.SetParameter("late", Lat(ppar->ye));

 return {ppar.release(), ""};
}

bool LayeredDataZ::Read(const MString& vname, std::map<MString, LayeredDataZ::Data>& cache, const BaseParameters* ip, size_t i) const
{
 if(cache.contains(vname)) return true;
 if(!isOk()) return false;

 auto p               = dynamic_cast<const struct Parameters*>(ip);
 auto [name, id, tid] = VarNameLoc(vname, times[i]);
 if(!name.Exist()) // Conversion read
  return TransformationRead(this, vname, cache, ip, i);

 // Direct read
 bool nodepth = false;
 Data data;
 //auto head = nc[id]->Header();
 for(const auto& v: nc[id].Vars())
  if(v.Name() == name)
  {
   if(v.NDim() == 3) nodepth = true;
   data = ReadVarRaw(nc[id], name, tid, nodepth, p);
   if(data)
   {
    cache[vname] = std::move(data);
    return true;
   }
  }

 return false;
}

LayeredDataZ::Data LayeredDataZ::ReadVarRaw(const NC& f, const MString& name, size_t i, bool nodepth, const struct LayeredDataZ::Parameters* p) const
{
 real unitmul = 1.0;
 real offset = 0.0, scale = 1.0;

 if(f.HasAtt(name, "add_offset")) offset = f.AttReal(name, "add_offset");
 if(f.HasAtt(name, "scale_factor")) scale = f.AttReal(name, "scale_factor");

 MString unit;
 if(f.HasAtt(name, "units")) unit = f.AttString(name, "units");
 if(unit == "m s-1" || unit == "m/s")
 {
  unitmul = 100.0;
  unit    = "cm/s";
 }

 Data data((p->xb < p->xe) ? (p->xe - p->xb + 1) : (dname.nx + p->xe - p->xb + 1), p->ye - p->yb + 1, Lon(p->xb), Lat(p->yb), lonstep, latstep, std::move(unit));

 auto trans = [scale, offset, unitmul](auto raw) -> real { return (raw * scale + offset) * unitmul; };

 auto rlayer = depthinv ? depths.size() - p->layer - 1 : p->layer;

 if(p->xb < p->xe)
 {
  auto ret = nodepth ? f.Read(name, data, trans, {dname.lonname, p->xb, p->xe - p->xb + 1}, {dname.latname, p->yb, p->ye - p->yb + 1}, {dname.timename, i, 1})
                     : f.Read(name, data, trans, {dname.lonname, p->xb, p->xe - p->xb + 1}, {dname.latname, p->yb, p->ye - p->yb + 1}, {dname.timename, i, 1},
                              {dname.depthname, rlayer, 1});
  if(!ret) return Data();
 }
 else
 {
  {
   auto ret = nodepth ? f.Read(name, data, trans, {dname.lonname, p->xb, dname.nx - p->xb + 1}, {dname.latname, p->yb, p->ye - p->yb + 1}, {dname.timename, i, 1})
                      : f.Read(name, data, trans, {dname.lonname, p->xb, dname.nx - p->xb + 1}, {dname.latname, p->yb, p->ye - p->yb + 1}, {dname.timename, i, 1},
                               {dname.depthname, rlayer, 1});
   if(!ret) return Data();
  }
  {
   size_t shift       = dname.nx - p->xb + 1;
   auto   shifteddata = [&data, shift](size_t ix, size_t iy) -> real& { return data(ix + shift, iy); };

   auto ret =
       nodepth ? f.Read(name, shifteddata, trans, {dname.lonname, 0, p->xe + 1}, {dname.latname, p->yb, p->ye - p->yb + 1}, {dname.timename, i, 1})
               : f.Read(name, shifteddata, trans, {dname.lonname, 0, p->xe + 1}, {dname.latname, p->yb, p->ye - p->yb + 1}, {dname.timename, i, 1}, {dname.depthname, rlayer, 1});
   if(!ret) return Data();
  }
 }
 return data;
}
