#define MICHLIB_NOSOURCE
#include "mirrorfuncs.h"
#include "StringFunctions.h"
#include "filehelpers.h"
#include "merrors.h"

using michlib::FD;
using michlib::message;

bool MakePath(const MString& dname)
{
 struct stat st;
 int         ret = stat(dname.Buf(), &st);
 if(ret == 0) return S_ISDIR(st.st_mode);

 auto    dirs = michlib::Split_on_words(dname, "/", false);
 MString cdir = "";
 for(const auto& dir: dirs)
 {
  cdir += "/" + dir;
  ret = stat(cdir.Buf(), &st);
  if(ret == 0 && S_ISDIR(st.st_mode)) continue;
  if(ret == 0 && !S_ISDIR(st.st_mode)) return false;
  ret = mkdir(cdir.Buf(), 0755);
  if(ret != 0) return false;
 }
 return true;
}

RetVal<std::vector<struct FileInfo>> ReadLocalFileList(const MString& dir, const MString& path)
{
 const static MString pref = "ReadLocalFileList";

 std::vector<struct FileInfo> out;
 DIRRAII                      dhandle;

 MakePath(dir);
 dhandle.reset(opendir(dir.Buf()));

 if(!dhandle) return {pref, "Can't open directory " + path + (path.Exist() ? "/" : "") + dir};

 int dfd             = dirfd(dhandle);
 errno               = 0;
 struct dirent* dent = readdir(dhandle);
 if(errno != 0) return {pref, "Can't read directory " + path + (path.Exist() ? "/" : "") + dir};
 struct stat st;

 do {
  if(dent->d_name[0] != '.')
  {
   int ret = fstatat(dfd, dent->d_name, &st, AT_SYMLINK_NOFOLLOW);
   if(ret != 0) return {pref, "Can't stat " + path + "/" + dir + "/" + dent->d_name};
   if(S_ISDIR(st.st_mode)) // Directory, recurse
   {
    auto list = ReadLocalFileList(dir + "/" + dent->d_name, path + (path.Exist() ? "/" : "") + dent->d_name);
    if(!list) return list;
    out.insert(out.end(), list.Value().begin(), list.Value().end());
   }
   if(S_ISREG(st.st_mode)) // Regular file
   {
    out.emplace_back(dir + "/" + dent->d_name, path + (path.Exist() ? "/" : "") + dent->d_name, MDateTime(st.st_mtim.tv_sec, st.st_mtim.tv_nsec), st.st_size);
   }
   // Ignore non-directories, non-files
  }
  dent = readdir(dhandle);
 } while(dent != nullptr || errno != 0);

 if(errno != 0) return {pref, "Can't read directory " + path + "/" + dir};
 std::sort(out.begin(), out.end(), [](const struct FileInfo& a, const struct FileInfo& b) { return a.name < b.name; });
 return out;
}

Error DownloadFile(const CURLRAII& chandle, const struct FileInfo& rinfo, const MString& root)
{
 const static MString pref = "DownloadFile";

 message("Downloading " + rinfo.url);

 MString dname = DirName(rinfo.name), fname = FileName(rinfo.name);
 FD      fd;

 if(!MakePath(root + "/" + dname)) return {pref, "Can't create directory " + root + "/" + dname};
 fd.Reset(creat((root + "/" + rinfo.name).Buf(), 0644));
 if(!fd) return {pref, "Can't create file " + root + "/" + rinfo.name};

 int cfd = fd.Get();
 curl_easy_setopt(chandle, CURLOPT_WRITEFUNCTION, Write2File);
 curl_easy_setopt(chandle, CURLOPT_WRITEDATA, &cfd);
 curl_easy_setopt(chandle, CURLOPT_URL, rinfo.url.Buf());
 auto res = curl_easy_perform(chandle);
 if(res != CURLE_OK)
 {
  unlink((root + "/" + rinfo.name).Buf());
  return {pref, MString("Can't download file: ") + chandle.Err()};
 }

 {
  struct timespec times[2];
  times[0].tv_sec = times[1].tv_sec = rinfo.mtime.Epoch();
  times[0].tv_nsec = times[1].tv_nsec = 0;

  int ret = futimens(fd, times);
  if(ret != 0)
  {
   unlink((root + "/" + rinfo.name).Buf());
   return {pref, "Can't set mtime for file: " + root + "/" + rinfo.name};
  }
 }

 return Error();
}

Error RemoveFile(const struct FileInfo& linfo)
{
 const static MString pref = "RemoveFile";

 message("Remove " + linfo.url);
 int ret = unlink(linfo.url.Buf());
 if(ret != 0) return {pref, "Can't remove file " + linfo.url};
 return Error();
}

Error UpdateFile(const CURLRAII& chandle, const struct FileInfo& rinfo, const struct FileInfo& linfo, const MString& root)
{
 const static MString pref = "UpdateFile";

 message("Update " + linfo.url);
 auto rm = RemoveFile(linfo);
 if(!rm) return rm.Add(pref, "Can't remove file");
 auto df = DownloadFile(chandle, rinfo, root);
 if(!df) return df.Add(pref, "Can't download file");

 return Error();
}
